import os.path
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import tensorflow as tf
import tensorflow_model_analysis as tfma
from absl import logging


@dataclass
class Entry:
    tp: int
    fp: int
    fn: int
    tn: int
    threshold: float


def load_entries(
    *,
    dir: str,
    model_name: str,
    slice_key: Optional[Dict[str, Any]] = None,
) -> List[Entry]:
    """
    Load confusion matrix statistics for multiple thresholds
    from `Evaluator` component output.
    """
    # Load evaluation step results
    result = tfma.load_eval_result(
        output_path=dir, output_file_format=None, model_name=model_name
    )
    if not result.model_location:
        # in case of pipeline running for the first time, there might be a single
        # unnamed model only in evaluation results. In such a case we will load
        # the data again without model name to retrieve the correct results.
        result = tfma.load_eval_result(output_path=dir, output_file_format=None)

    # Select data for given slice
    if slice_key is None:
        slice_spec = tfma.slicer.slicer_lib.SingleSliceSpec()
    else:
        slice_spec = tfma.SlicingSpec(feature_values=slice_key)
    data, config = tfma.view.util.get_plot_data_and_config(
        result.plots, slice_spec
    )
    # Extract confusion matrix data
    confusion_matrix_config = config["metricKeys"]["confusionMatrixPlot"]
    plot_data = data[confusion_matrix_config["metricName"]][
        confusion_matrix_config["dataSeries"]
    ]
    return [
        Entry(
            threshold=entry.get("threshold", 0.0),
            tp=int(entry.get("truePositives", 0.0)),
            fp=int(entry.get("falsePositives", 0.0)),
            fn=int(entry.get("falseNegatives", 0.0)),
            tn=int(entry.get("trueNegatives", 0.0)),
        )
        for entry in plot_data
    ]


def g_mean(entry: Entry) -> float:
    sensitivity = entry.tp / (entry.tp + entry.fn)
    specificity = entry.tn / (entry.fp + entry.tn)
    return (sensitivity * specificity) ** 2


def find_best_binary_classification_threshold(
    *, dir: str, model_name: str
) -> float:
    entries = load_entries(dir=dir, model_name=model_name, slice_key=None)
    entries.sort(key=g_mean, reverse=True)
    best_entry, *_rest = entries
    logging.debug(
        "Best threshold found %s with G-mean %f",
        best_entry,
        g_mean(best_entry),
    )
    return best_entry.threshold


def load_best_threshold(dir: str) -> float:
    file_path = os.path.join(dir, "value")
    with tf.io.gfile.GFile(file_path, "r") as f:
        return float(f.read())



import os.path
from dataclasses import dataclass
from typing import List

import tensorflow as tf
from absl import logging
from typing_extensions import Protocol


class NewestPathGetterInterface(Protocol):
    """Interface for tools for getting newest pipeline artifacts directory."""

    def __call__(self, component_name: str, output_name: str) -> str:
        """
        Get path to newest artifact.

        `component_name` is the name of the component,
        `output_name` is the name of component output.
        """


def get_newest_local_subdirectory(directory: str) -> str:
    subdirectories: List[str] = tf.io.gfile.glob(os.path.join(directory, "*"))
    newest_subdir = max(
        subdirectories,
        key=lambda subdirectory: int(os.path.basename(subdirectory)),
    )
    logging.debug(
        "From %d subdirs selected %r as newest",
        len(subdirectories),
        newest_subdir,
    )
    return newest_subdir


@dataclass
class NewestLocalPathGetter:
    """
    Tool for obtaining paths to newest artifact directories for your components.

    When running locally each components creates a directory for artifacts with
    sequential integer name. This tool helps in obtaining the most recent one
    (that might not be what you want if you component failed).
    """

    artifact_dir: str
    """
    Path to pipeline output directory.
    """

    pipeline_name: str
    """
    Name of the pipeline.
    """

    def __call__(self, component_name: str, output_name: str) -> str:
        path = os.path.join(
            self.artifact_dir, self.pipeline_name, component_name, output_name
        )
        return get_newest_local_subdirectory(path)


@dataclass
class NewestVertexAIPathGetter:
    """
    Tool for obtaining paths to newest artifact directories for your components.

    When running on VertexAI each run creates a new directory with epoch timestamp name.
    Within that directory a new directory with pipeline name prefix is created.
    Inside it there are directories for components, each prefixed with component name.
    Within the component directory there is one directory for each output.

    This tool helps in obtaining the most recent output directory
    for given component's output
    (that might not be what you want if you component failed
    or if you have multiple pipeline jobs running simultainously).
    """

    artifact_dir: str
    """
    Path to pipeline output directory on Google Cloud Storage.
    """
    pipeline_name: str
    """
    Name of the pipeline.
    """

    def __call__(self, component_name: str, output_name: str) -> str:
        pipeline_run_dirs = tf.io.gfile.glob(
            os.path.join(self.artifact_dir, self.pipeline_name, "*")
        )
        newest_run_dir = max(
            pipeline_run_dirs,
            key=lambda subdirectory: int(os.path.basename(subdirectory)),
        )
        logging.debug(
            "From %d subdirs selected %r as newest run dir",
            len(pipeline_run_dirs),
            newest_run_dir,
        )
        dirs = tf.io.gfile.glob(os.path.join(newest_run_dir, "*"))
        assert len(dirs) == 1, f"Expected a single subdirectory, but got {len(dirs)}"
        (subdir,) = dirs
        assert os.path.basename(subdir).startswith(
            self.pipeline_name
        ), f"Expected subdirectory to include pipeline name ({self.pipeline_name!r}) but got {subdir!r}"
        component_dirs = tf.io.gfile.glob(os.path.join(subdir, f"{component_name}_*"))
        assert (
            len(dirs) == 1
        ), f"Expected a single component directory, but got {len(component_dirs)}"
        (component_dir,) = component_dirs
        result = os.path.join(component_dir, output_name)
        logging.debug("Selected %r as newest", result)
        return result



import os.path
from typing import Any, Iterable, Mapping, Optional, Tuple

import tfx.v1 as tfx
from absl import logging
from ml_metadata.proto import metadata_store_pb2
from tfx.dsl.components.base.base_component import BaseComponent
from tfx.types.channel import Channel

from .base import BasePipelineHelper
from .interface import DEFAULT_CUSTOM_CONFIG, Resources


class LocalPipelineHelper(BasePipelineHelper, arbitrary_types_allowed=True):
    model_push_destination: tfx.proto.PushDestination

    def construct_trainer(
        self,
        *,
        examples: Optional[Channel] = None,
        transform_graph: Optional[Channel] = None,
        schema: Optional[Channel] = None,
        base_model: Optional[Channel] = None,
        hyperparameters: Optional[Channel] = None,
        run_fn: str,
        train_args: Optional[tfx.proto.TrainArgs] = None,
        eval_args: Optional[tfx.proto.EvalArgs] = None,
        custom_config: Mapping[str, Any] = DEFAULT_CUSTOM_CONFIG,
    ) -> BaseComponent:
        return tfx.components.Trainer(
            examples=examples,
            transform_graph=transform_graph,
            schema=schema,
            base_model=base_model,
            hyperparameters=hyperparameters,
            run_fn=run_fn,
            train_args=train_args,
            eval_args=eval_args,
            custom_config=dict(custom_config),
        )

    def construct_tuner(
        self,
        *,
        examples: Channel,
        schema: Optional[Channel] = None,
        transform_graph: Optional[Channel] = None,
        base_model: Optional[Channel] = None,
        tuner_fn: str,
        train_args: Optional[tfx.proto.TrainArgs] = None,
        eval_args: Optional[tfx.proto.EvalArgs] = None,
        custom_config: Mapping[str, Any] = DEFAULT_CUSTOM_CONFIG,
    ) -> BaseComponent:
        return tfx.components.Tuner(
            examples=examples,
            schema=schema,
            transform_graph=transform_graph,
            base_model=base_model,
            tuner_fn=tuner_fn,
            train_args=train_args,
            eval_args=eval_args,
            custom_config=dict(custom_config),
        )

    def construct_pusher(
        self,
        *,
        model: Optional[Channel] = None,
        model_blessing: Optional[Channel] = None,
        infra_blessing: Optional[Channel] = None,
        custom_config: Mapping[str, Any] = DEFAULT_CUSTOM_CONFIG,
    ) -> BaseComponent:
        return tfx.components.Pusher(
            model=model,
            model_blessing=model_blessing,
            infra_blessing=infra_blessing,
            push_destination=self.model_push_destination,
            custom_config=dict(custom_config),
        )

    def get_metadata_connection_config(self) -> metadata_store_pb2.ConnectionConfig:
        metadata_path = os.path.join(
            self.output_dir, "tfx_metadata", self.pipeline_name, "metadata.db"
        )
        logging.info("Pipeline will store metadata in %r", metadata_path)
        return tfx.orchestration.metadata.sqlite_metadata_connection_config(
            metadata_path
        )

    def create_and_run_pipeline(
        self,
        components: Iterable[BaseComponent],
        enable_cache: bool = False,
    ) -> None:
        logging.info(
            "Creating local pipeline name=%r, root=%r, enable_cache=%r",
            self.pipeline_name,
            self.pipeline_root,
            enable_cache,
        )
        metadata_connection_config = self.get_metadata_connection_config()
        pipeline = tfx.dsl.Pipeline(
            pipeline_name=self.pipeline_name,
            pipeline_root=self.pipeline_root,
            components=list(components),
            enable_cache=enable_cache,
            metadata_connection_config=metadata_connection_config,
        )

        logging.info("Runnig pipeline using local DAG runner")
        tfx.orchestration.LocalDagRunner().run(pipeline)
        logging.info("Pipeline run finished")



from typing import Type, TypeVar

import tensorflow as tf
from google.protobuf.message import Message

ProtoClass = TypeVar("ProtoClass", bound=Message)


def read_binary_proto(file_path: str, proto_class: Type[ProtoClass]) -> ProtoClass:
    instance: ProtoClass = proto_class()
    with tf.io.gfile.GFile(file_path, "rb") as f:
        data = f.read()
    instance.ParseFromString(data)
    return instance



from typing import Any, Dict, List, Optional, Tuple

import tensorflow_model_analysis as tfma

# In Evaluator component usually you will want to compare the `candidate` model
# against a `baseline` model. You are more likely to want to view the stats
# of the `candidate` model, as it's the one you are training.
DEFAULT_MODEL_NAME = "candidate"


def passed_validation(dir: str) -> bool:
    """
    Check if model passed validation.

    Consumes `Evaluator` output.
    """
    validation_result = tfma.load_validation_result(dir)
    validation_passed: bool = validation_result.validation_ok
    return validation_passed


def get_slice_names(
    dir: str, model_name: str = DEFAULT_MODEL_NAME
) -> List[Tuple[Tuple[str, Any], ...]]:
    """
    Return slice names that are available for analysis.

    Consumes `Evaluator` output.
    """
    result = tfma.load_eval_result(
        output_path=dir, output_file_format=None, model_name=model_name
    )
    slice_names: List[Tuple[Tuple[str, Any], ...]] = result.get_slice_names()
    return slice_names


def display_metrics(
    dir: str,
    model_name: str = DEFAULT_MODEL_NAME,
    column: Optional[str] = None,
) -> Any:
    """
    Display metrics inside a notebook cell.

    Provide `column` if you configured `slicing_specs` in `EvalConfig`
    and want to see metrics breakdown by slice.

    Consumes `Evaluator` output.
    """
    result = tfma.load_eval_result(
        output_path=dir, model_name=model_name, output_file_format=None
    )
    return tfma.view.render_slicing_metrics(result, slicing_column=column)


def display_plots(
    dir: str,
    model_name: str = DEFAULT_MODEL_NAME,
    slice_key: Optional[Dict[str, Any]] = None,
) -> Any:
    """
    Display plots inside a notebook cell.

    Provide `slice_key` if you want to see plots for a particular slice
    of the dataset.

    Consumes `Evaluator` output.
    """
    result = tfma.load_eval_result(
        output_path=dir, model_name=model_name, output_file_format=None
    )
    if slice_key:
        slicing_spec = tfma.SlicingSpec(feature_values=slice_key)
    else:
        slicing_spec = None
    return tfma.view.render_plot(result, slicing_spec)



import os.path
from typing import Any

import tensorflow_data_validation as tfdv
import tensorflow_metadata as tfmd

from .utils import read_binary_proto


def display_anomalies(dir: str, split_name: str = "all") -> Any:
    """
    Display dataset anomalies in a notebook cell.

    Consumes ExampleValidator output.
    """
    path = os.path.join(dir, f"Split-{split_name}", "SchemaDiff.pb")
    anomalies = read_binary_proto(path, tfmd.proto.anomalies_pb2.Anomalies)
    return tfdv.display_anomalies(anomalies)
